from typing import Any
from typing import List
from typing import Dict


def list_modules(
    module_names: List[str] = [],
    module_types: List[str] = [],
    tags: List[str] = [],
    search_store: bool = False,
    nameonly: bool = False,
    humanized_size: bool = False,
) -> List[Dict[str, Any]]:
    import re
    from ...lib.module.remote import search_remote
    from ...lib.module.local import search_local
    from ...lib.module.local import get_local_module_info
    from ...lib.module.remote import get_remote_module_info_ls
    from ...lib.util.util import humanize_bytes
    from ...lib.module.remote import RemoteModuleLs

    all_toks_json = []
    if search_store:
        if module_types:
            module_names_to_search = []
            for mt in module_types:
                module_names_to_search.extend(
                    search_remote(*module_names, module_type=mt)
                )
        else:
            module_names_to_search = search_remote(*module_names)
    else:
        module_names_to_search = search_local(*module_names)
    if not module_names_to_search:
        return all_toks_json
    for module_name in module_names_to_search:
        if search_store:
            module_info = get_remote_module_info_ls(module_name)
            if module_info:
                add_local_module_info_to_remote_module_info(module_info)
        else:
            module_info = get_local_module_info(module_name)
        if not module_info:
            continue
        if module_types and module_info.type not in module_types:
            continue
        if tags and module_info.tags:
            if not module_info.tags:
                continue
            matched = False
            for pattern in tags:
                for tag in module_info.tags:
                    if re.match(pattern, tag):
                        matched = True
                        break
                if matched:
                    break
            if not matched:
                continue
        if isinstance(module_info, RemoteModuleLs):
            size = module_info.size
        else:
            size = module_info.get_size()
        if humanized_size:
            size = humanize_bytes(size)
        toks = {"name": module_name}
        if not nameonly:
            if search_store:
                toks.update(
                    {
                        "title": module_info.title,
                        "type": module_info.type,
                        "size": size,
                        "version": module_info.latest_code_version,
                        "data_source": module_info.latest_data_source,
                        "installed": module_info.installed,
                        "local_code_version": module_info.local_code_version,
                        "local_data_source": module_info.local_data_source,
                    }
                )
            else:
                toks.update(
                    {
                        "title": module_info.title,
                        "type": module_info.type,
                        "size": size,
                        "version": module_info.latest_code_version,
                        "data_source": module_info.latest_data_source,
                    }
                )
        all_toks_json.append(toks)
    return all_toks_json


def add_local_module_info_to_remote_module_info(remote_info):
    from ...lib.module.local import get_local_module_info

    local_info = get_local_module_info(remote_info.name)
    if local_info:
        remote_info.installed = "yes"
        remote_info.local_code_version = local_info.latest_code_version
        remote_info.local_data_source = local_info.latest_data_source
    else:
        remote_info.installed = ""
        remote_info.local_code_version = ""
        remote_info.local_data_source = ""



DEFAULT_GUI_PORT = 8080
default_gui_port_ssl = 8443
DEFAULT_JOB_TABLE_PAGESIZE = 10
job_table_pagesize_key = "job_table_pagesize"
DEFAULT_RESULT_VIEWER_NUM_VAR_LIMIT_FOR_GENE_SUMMARY = 100000
result_viewer_num_var_limit_for_gene_summary_key = (
    "result_viewer_num_var_limit_for_gene_summary"
)
DEFAULT_RESULT_VIEWER_NUM_VAR_LIMIT_FOR_SUMMARY_WIDGET = 1000
result_viewer_num_var_limit_for_summary_widget_key = (
    "result_viewer_num_var_limit_for_summary_widget"
)
LOG_FN = "gui.log"
DEFAULT_PRIVATE_KEY = "default_private_key"
COOKIE_KEY = "oakvar_token"
WS_COOKIE_KEY = "ws_id"
SYSTEM_STATE_CONNECTION_KEY = "connection"
SYSTEM_STATE_SETUP_KEY = "setup"
SYSTEM_STATE_INSTALL_KEY = "install"
SYSTEM_STATE_INSTALL_QUEUE_KEY = "install_queue"
SYSTEM_MSG_KEY = "msg_kind"
SYSTEM_MESSAGE_DB_FNAME = "system_messages.sqlite"
SYSTEM_MESSAGE_TABLE = "system_messages"
SYSTEM_ERROR_TABLE = "system_errors"
INSTALL_KILL_SIGNAL = "kill_signal"
PORT_KEY = "port"
SYSCONF_PORT_KEY = "gui_port"
SYSCONF_SSL_PORT_KEY = "gui_port_ssl"
SYSCONF_HOST_KEY = "gui_host"
SYSCONF_SSL_HOST_KEY = "gui_host_ssl"
SSL_ENABELD_KEY = "ssl_enabled"



class WebSocketHandlers:
    def __init__(
        self, system_worker_state=None, wss={}, system_message_last_ids={}, logger=None
    ):
        from .system_message_db import get_system_message_db_conn

        self.routes = []
        self.system_worker_state = system_worker_state
        self.wss = wss
        self.system_message_last_ids = system_message_last_ids
        self.logger = logger
        self.conn = get_system_message_db_conn()
        self.cursor = self.conn.cursor()
        self.add_routes()

    def add_routes(self):
        self.routes = []
        self.routes.append(["GET", "/ws", self.connect])

    async def connect(self, request):
        import asyncio
        from aiohttp.web import WebSocketResponse
        import concurrent.futures
        from uuid import uuid4
        from .consts import WS_COOKIE_KEY
        from .consts import SYSTEM_STATE_CONNECTION_KEY
        from .consts import SYSTEM_MSG_KEY
        from .system_message_db import get_last_msg_id

        assert self.system_worker_state is not None
        ws_id = request.cookies.get(WS_COOKIE_KEY)
        if ws_id and ws_id in self.wss:
            del self.wss[ws_id]
        ws_id = str(uuid4())
        ws = WebSocketResponse(timeout=60 * 60 * 24 * 365)
        self.wss[ws_id] = ws
        await ws.prepare(request)
        try:
            await ws.send_json(
                {SYSTEM_MSG_KEY: SYSTEM_STATE_CONNECTION_KEY, WS_COOKIE_KEY: ws_id}
            )
        except ConnectionResetError:
            raise
        except Exception:
            raise
        to_dels = []
        for ws_id in self.wss:
            ws_t = self.wss[ws_id]
            if ws_t.closed:
                to_dels.append(ws_id)
        for ws_id in to_dels:
            del self.wss[ws_id]
        last_msg_id = get_last_msg_id(self.conn)
        while True:
            try:
                await asyncio.sleep(1)
                if ws.closed:
                    break
                last_msg_id = await self.process_system_worker_state(
                    ws=ws, last_msg_id=last_msg_id
                )
            except concurrent.futures._base.CancelledError:
                pass
            except ConnectionResetError:
                break
            except Exception as e:
                if self.logger:
                    self.logger.exception(e)
                break
        return ws

    async def process_setup_state(self, ws=None, last_msg_id=0):
        from .consts import SYSTEM_STATE_SETUP_KEY
        from .consts import SYSTEM_STATE_INSTALL_KEY
        from .consts import SYSTEM_MSG_KEY
        from .consts import SYSTEM_MESSAGE_TABLE

        if ws is None or not self.system_worker_state:
            return last_msg_id
        self.cursor.execute(
            f"select uid, kind, msg, dt from {SYSTEM_MESSAGE_TABLE} where uid > ?",
            (last_msg_id,),
        )
        ret = self.cursor.fetchall()
        if ret:
            last_msg_id = max([v[0] for v in ret])
            setup_items = []
            install_items = []
            for row in ret:
                kind = row[1]
                if kind == "setup":
                    setup_items.append({"kind": kind, "msg": row[2], "dt": row[3]})
                elif kind == "install":
                    install_items.append({"kind": kind, "msg": row[2], "dt": row[3]})
                else:
                    setup_items.append({"kind": kind, "msg": row[2], "dt": row[3]})
            if setup_items:
                await ws.send_json(
                    {SYSTEM_MSG_KEY: SYSTEM_STATE_SETUP_KEY, "items": setup_items}
                )
            if install_items:
                await ws.send_json(
                    {SYSTEM_MSG_KEY: SYSTEM_STATE_INSTALL_KEY, "items": install_items}
                )
        return last_msg_id

    async def process_system_worker_state(self, ws=None, last_msg_id=0):
        if ws is None:
            return last_msg_id
        last_msg_id = await self.process_setup_state(ws=ws, last_msg_id=last_msg_id)
        return last_msg_id



loop = None


def get_event_loop():
    # from sys import platform as sysplatform
    import asyncio
    import nest_asyncio

    # import uvloop

    # uvloop.install()
    # asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

    global loop
    if not loop:
        # if sysplatform == "win32":  # Required to use asyncio subprocesses
        #    #from asyncio import ProactorEventLoop  # type: ignore
        #    import asyncio
        #    #loop = ProactorEventLoop()
        #    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
        #    loop = get_event_loop()
        # else:
        #    loop = get_event_loop()
        loop = asyncio.get_event_loop()
        nest_asyncio.apply(loop)
    return loop



from oakvar import BaseReporter
from typing import Union
from typing import Any
from typing import Dict
from typing import List


class Reporter(BaseReporter):
    """Reporter module template."""

    def write_preface(self, level: str):
        """Method to set up output files and write "preface".
        Preface means content before the headers for output columns.
        Opening an output file and saving the file handler should be done
        in this method.

        Args:
            level (str): Level of reporting. For example, "variant" or "gene".
        """
        _ = level
        pass

    def write_header(self, level: str):
        """Method to write column headers.
        This method is supposed to write column headers.
        The file handler saved in write_preface method is supposed to be
        used here to write column headers.

        Args:
            level (str): Level of reporting. For example, "variant" or "gene".
        """
        _ = level
        pass

    def write_table_row(self, row: Union[Dict[str, Any], List[Any]]):
        """Method to write annotation results.

        The file handler saved in write_preface method is supposed to be
        used here to write column headers.

        Variants are sent to this method one by one as `row`. It is either
        a dictionary or a list. If `self.dictrow` is `True`, a dict will be
        passed to this method. If not, a list will. The dict's keys will be
        output column names. If a list is passed, `self.columns` will have
        the column information in the same order as in the list.

        Args:
            row (Union[Dict[str, Any], List[Any]]): annotation result for a variant.
        """
        _ = row
        pass



from typing import Optional
from ...lib.module import InstallProgressHandler


class InstallProgressStdout(InstallProgressHandler):
    def __init__(
        self,
        module_name: Optional[str] = None,
        module_version: Optional[str] = None,
        outer=None,
    ):
        super().__init__(
            module_name=module_name, module_version=module_version, outer=outer
        )
        self.system_worker_state = None

    def stage_start(self, stage):
        self.cur_stage = stage
        if self.outer:
            self.outer.write(self._stage_msg(stage))



from . import cli_entry
from . import cli_func


@cli_entry
def cli_update(args):
    return update(args)


@cli_func
def update(args, __name__="update"):
    from ..api import update

    return update(**args)


def get_parser_ov_update():
    from argparse import ArgumentParser

    parser_ov_update = ArgumentParser(
        description="Updates OakVar to the latest version.",
    )
    parser_ov_update.set_defaults(func=cli_update)
    return parser_ov_update



from typing import Union
from typing import Optional
from pathlib import Path


class BaseFile(object):
    valid_types = ["string", "int", "float"]

    def __init__(self, path):
        from os.path import abspath

        self.path = abspath(path)
        self.columns = {}

    def _validate_col_type(self, col_type):
        if col_type not in self.valid_types:
            raise Exception(
                "Invalid column type {} in {}. Choose from {}".format(
                    col_type, self.path, ", ".join(self.valid_types)
                )
            )

    def get_col_def(self, col_index):
        return self.columns[col_index]

    def get_all_col_defs(self):
        return self.columns


class FileReader(BaseFile):
    def __init__(self, path, seekpos: int=0, chunksize: Optional[int]=None, logger=None):
        from .util import detect_encoding

        super().__init__(path)
        self.seekpos = seekpos
        self.chunksize = chunksize
        self.encoding = detect_encoding(self.path)
        self.annotator_name = ""
        self.annotator_displayname = ""
        self.annotator_version = ""
        self.index_columns = []
        self.report_substitution = None
        self.f = None
        self.csvfmt: bool = False
        self.logger = logger
        self._setup_definition()

    def _setup_definition(self):
        from json import loads
        from json.decoder import JSONDecodeError

        with open(self.path, encoding=self.encoding) as f:
            line = f.readline()[:-1]
            if line.startswith("#fmt=csv"):
                self.csvfmt = True
        for line in self._loop_definition():
            if line.startswith("#name="):
                self.annotator_name = line.split("=")[1]
            elif line.startswith("#displayname="):
                self.annotator_displayname = line.split("=")[1]
            elif line.startswith("#version="):
                self.annotator_version = line.split("=")[1]
            elif line.startswith("#index="):
                cols = line.split("=")[1].split(",")
                self.index_columns.append(cols)
            elif line.startswith("#column="):
                coldef = ColumnDefinition({})
                col_s = "=".join(line.split("=")[1:])
                try:
                    coldef.from_json(col_s)
                except JSONDecodeError:
                    if self.logger:
                        self.logger.error(f"column definition error: {col_s}")
                    coldef.from_var_csv(col_s)
                self._validate_col_type(coldef.type)
                self.columns[coldef.index] = coldef
            elif line.startswith("#report_substitution="):
                self.report_substitution = loads(line.split("=")[1])
            else:
                continue

    def get_index_columns(self):
        return self.index_columns

    def override_column(
        self, index, name, title=None, data_type="string", cats=[], category=None
    ):
        if title is None:
            title = " ".join(x.title() for x in name.split("_"))
        if index not in self.columns:
            self.columns[index] = ColumnDefinition({})
        self.columns[index].title = title
        self.columns[index].name = name
        self.columns[index].type = data_type
        self.columns[index].categories = cats
        self.columns[index].category = category

    def get_column_names(self):
        sorted_order = sorted(list(self.columns.keys()))
        return [self.columns[x].name for x in sorted_order]

    def get_annotator_name(self):
        return self.annotator_name

    def get_annotator_displayname(self):
        return self.annotator_displayname

    def get_annotator_version(self):
        return self.annotator_version

    def get_chunksize(self, num_core):
        f = open(self.path)
        max_data_line_no = 0
        max_line_no = 0
        for line in f:
            max_line_no += 1
            if line.startswith("#"):
                continue
            max_data_line_no += 1
        chunksize = max(int(max_data_line_no / num_core), 1)
        f.seek(0)
        poss = [[0, 0]]
        line_no = 0
        data_line_no = 0
        while True:
            line = f.readline()
            line_no += 1
            if not line.startswith("#"):
                data_line_no += 1
                if data_line_no % chunksize == 0 and len(poss) < num_core:
                    poss.append([f.tell(), chunksize])
            if line_no == max_line_no:
                break
        f.close()
        len_poss = len(poss)
        return max_line_no, chunksize, poss, len_poss, max_data_line_no

    def loop_data(self):
        from ..exceptions import BadFormatError
        from json import loads

        for lnum, toks in self._loop_data():
            out = {}
            if len(toks) < len(self.columns):
                err_msg = "Too few columns. Received %s. Expected %s." % (
                    len(toks),
                    len(self.columns),
                )
                return BadFormatError(err_msg)
            for col_index, col_def in self.columns.items():
                col_name = col_def.name
                col_type = col_def.type
                tok = toks[col_index]
                if tok == "":
                    out[col_name] = None
                else:
                    if col_type == "string":
                        out[col_name] = tok
                    elif col_type == "int":
                        try:
                            out[col_name] = int(tok)
                        except ValueError:
                            try:
                                out[col_name] = int(float(tok))
                            except Exception:
                                out[col_name] = None
                        except Exception:
                            out[col_name] = None
                    elif col_type == "float":
                        try:
                            tok = loads(tok)
                            if type(tok) == list:
                                out[col_name] = ",".join([str(v) for v in tok])
                            else:
                                out[col_name] = float(tok)
                        except Exception:
                            tok = None
            yield lnum, toks, out

    def get_data(self):
        all_data = [d for _, _, d in self.loop_data()]
        return all_data

    def _loop_definition(self):
        if self.csvfmt:
            f = open(self.path, newline="", encoding=self.encoding)
        else:
            f = open(self.path, encoding=self.encoding)
        for line in f:
            line = line.rstrip().lstrip()
            if line.startswith("#"):
                yield line
            else:
                break
        f.close()

    def _loop_data(self):
        if not self.encoding:
            return
        if self.csvfmt:
            with open(self.path, newline="") as f:
                f.seek(self.seekpos)
                lnum = 0
                import csv

                csv.field_size_limit(
                    1147483647
                )  # crude way. 2147483647 is the min of C long max.
                csvreader = csv.reader(f)
                num_data_rows = 0
                for row in csvreader:
                    if row[0].startswith("#"):
                        continue
                    yield csvreader.line_num, row
                    num_data_rows += 1
                    if self.chunksize and num_data_rows >= self.chunksize:
                        return
        else:
            with open(self.path, "rb") as f:
                if self.seekpos is not None:
                    f.seek(self.seekpos)
                lnum = 0
                for line in f:
                    line = line.decode(self.encoding)
                    if line.startswith("#"):
                        continue
                    else:
                        line = line.rstrip("\r\n")
                        yield lnum, line.split("\t")
                    lnum += 1
                    if self.chunksize is not None and lnum == self.chunksize:
                        break


class FileWriter(BaseFile):
    def __init__(
        self,
        path,
        include_definition=True,
        include_titles=True,
        titles_prefix="#",
        columns=[],
        mode="w",
        fmt="csv",
    ):
        super().__init__(path)
        self.csvfmt: bool = False
        if fmt == "csv":
            self.csvfmt = True
        self.csvwriter = None
        if fmt == "csv":
            self.wf = open(self.path, mode, newline="", encoding="utf-8")
            from csv import writer

            self.csvwriter = writer(self.wf)
            if mode == "w":
                self.wf.write("#fmt=csv\n")
        else:
            self.wf = open(self.path, mode, encoding="utf-8")
        self.mode: str = mode
        self.ready_to_write = False
        self.ordered_columns = []
        self.name_to_col_index = {}
        self.title_toks = []
        self.include_definition = include_definition
        self.include_titles = include_titles
        self.titles_prefix = titles_prefix
        self.add_columns(columns)

    def add_column(self, col_d):
        col_index = len(self.columns)
        col_d["index"] = col_index
        col_def = ColumnDefinition(col_d)
        for i in self.columns:
            if self.columns[i].name == col_def.name:
                continue
        self.columns[col_index] = col_def

    def add_columns(self, col_defs):
        for col_def in col_defs:
            self.add_column(col_def)

    def prep_for_write(self):
        if self.ready_to_write:
            return
        col_indices = sorted(self.columns.keys())
        correct_index = -1
        for col_index in col_indices:
            correct_index += 1
            if correct_index != col_index:
                raise Exception("Column %d must be defined" % correct_index)
            col_def = self.columns[col_index]
            self.ordered_columns.append(col_def)
            self.title_toks.append(col_def.title)
            self.name_to_col_index[col_def.name] = col_index
        self.ready_to_write = True

    def write_names(self, annotator_name, annotator_display_name, annotator_version):
        line = "#name={:}\n".format(annotator_name)
        self.wf.write(line)
        line = "#displayname={:}\n".format(annotator_display_name)
        self.wf.write(line)
        line = "#version={:}\n".format(annotator_version)
        self.wf.write(line)
        self.wf.flush()

    def add_index(self, index_columns):
        self.write_meta_line("index", ",".join(index_columns))

    def write_meta_line(self, key, value):
        line = "#{:}={:}\n".format(key, value)
        self.wf.write(line)
        self.wf.flush()

    def write_definition(self, conf=None):
        from json import dumps

        self.prep_for_write()
        for col_def in self.ordered_columns:
            self.write_meta_line("column", col_def.get_json())
        if conf and "report_substitution" in conf:
            self.write_meta_line(
                "report_substitution", dumps(conf["report_substitution"])
            )
        self.wf.flush()

    def write_input_paths(self, input_path_dict):
        from json import dumps

        s = "#input_paths={}\n".format(dumps(input_path_dict))
        self.wf.write(s)
        self.wf.flush()

    def write_data(self, data):
        if not data:
            return
        self.prep_for_write()
        wtoks = [data.get(col.name, "") for col in self.columns.values()]
        if self.csvfmt:
            if self.csvwriter is not None:
                try:
                    self.csvwriter.writerow(wtoks)
                except Exception:
                    import traceback

                    traceback.print_exc()
        else:
            self.wf.write("\t".join(wtoks) + "\n")

    def close(self):
        self.wf.close()


class CrxMapping(object):
    def __init__(self):
        from re import compile
        from typing import Optional

        self.protein: Optional[str] = None
        self.achange = None
        self.transcript: Optional[str] = None
        self.tchange = None
        self.so: Optional[str] = None
        self.gene = None
        self.tref = None
        self.tpos_start = None
        self.talt = None
        self.aref = None
        self.apos_start = None
        self.aalt = None
        self.mapping = None
        self.tchange_re = compile(r"([AaTtCcGgUuNn_-]+)(\d+)([AaTtCcGgUuNn_-]+)")
        self.achange_re = compile(r"([a-zA-Z_\*]+)(\d+)([AaTtCcGgUuNn_\*]+)")

    def load_tchange(self, tchange):
        self.tchange = tchange
        if tchange is not None:
            self.parse_tchange()

    def parse_tchange(self):
        if self.tchange is not None:
            tchange_match = self.tchange_re.match(self.tchange)
            if tchange_match:
                self.tref = tchange_match.group(1)
                self.tpos_start = int(tchange_match.group(2))
                self.talt = tchange_match.group(3)

    def load_achange(self, achange):
        self.achange = achange
        if self.achange is not None:
            self.parse_achange()

    def parse_achange(self):
        if self.achange is not None:
            achange_match = self.achange_re.match(self.achange)
            if achange_match:
                self.aref = achange_match.group(1)
                self.apos_start = int(achange_match.group(2))
                self.aalt = achange_match.group(3)


class AllMappingsParser(object):
    def __init__(self, s):
        from json import loads
        from collections import OrderedDict

        if type(s) == str:
            self._d = loads(s, object_pairs_hook=OrderedDict)
        else:
            self._d = s
        self._protein_index = 0
        self._achange_index = 1
        self._so_index = 2
        self._transc_index = 3
        self._tchange_index = 4
        self.mappings = self.get_all_mappings()

    def get_genes(self):
        return list(self._d.keys())

    def get_uniq_sos(self):
        sos = {}
        for mapping in self.mappings:
            for so in mapping.so.split(","):
                sos[so] = True
        sos = list(sos.keys())
        return sos

    def none_to_empty(self, s):
        if s is None:
            return ""
        else:
            return s

    def get_mapping(self, t):
        mapping = CrxMapping()
        mapping.transcript = self.none_to_empty(t[self._transc_index])
        mapping.so = self.none_to_empty(t[self._so_index])
        mapping.load_tchange(self.none_to_empty(t[self._tchange_index]))
        mapping.load_achange(self.none_to_empty(t[self._achange_index]))
        mapping.protein = self.none_to_empty(t[self._protein_index])
        return mapping

    def get_all_mappings(self):
        mappings = []
        for gene, ts in self._d.items():
            for t in ts:
                mapping = self.get_mapping(t)
                mapping.gene = gene
                mappings.append(mapping)
        return mappings

    def get_transcript_mapping(self, transcript):
        for mapping in self.mappings:
            if mapping.transcript == transcript:
                return mapping
        return None


class ColumnDefinition(object):

    csv_order = [
        "index",
        "title",
        "name",
        "type",
        "categories",
        "width",
        "desc",
        "hidden",
        "category",
        "filterable",
        "hide_from_gui_filter",
        "link_format",
        "genesummary",
    ]

    db_order = [  # TODO change name to denote legacy
        "col_name",
        "col_title",
        "col_type",
        "col_cats",
        "col_width",
        "col_desc",
        "col_hidden",
        "col_ctg",
        "col_filterable",
        "col_hide_from_gui_filter",
        "col_link_format",
    ]

    sql_map = {
        "col_name": "name",
        "col_title": "title",
        "col_type": "type",
        "col_cats": "categories",
        "col_width": "width",
        "col_desc": "desc",
        "col_hidden": "hidden",
        "col_ctg": "category",
        "col_filterable": "filterable",
        "col_hide_from_gui_filter": "hide_from_gui_filter",
        "col_link_format": "link_format",
        "col_genesummary": "genesummary",
    }

    def __init__(self, d):
        self.index = None
        self.name = None
        self.title = None
        self.type = None
        self.categories = None
        self.width = None
        self.desc = None
        self.hidden = None
        self.category = None
        self.filterable = None
        self.hide_from_gui_filter = None
        self.link_format = None
        self.genesummary = None
        self.table = None
        self.level = None
        self._load_dict(d)

    def _load_dict(self, d):
        self.index = d.get("index")
        self.name = d.get("name")
        self.title = d.get("title")
        self.type = d.get("type")
        self.categories = d.get("categories", [])
        self.width = d.get("width")
        self.desc = d.get("desc")
        self.hidden = bool(d.get("hidden", False))
        self.category = d.get("category")
        self.filterable = bool(d.get("filterable", True))
        self.hide_from_gui_filter = bool(d.get("hide_from_gui_filter", False))
        self.link_format = d.get("link_format")
        self.genesummary = d.get("genesummary", False)
        self.table = d.get("table", False)
        self.level = d.get("level")

    def from_row(self, row, order=None):
        from json import loads

        if order is None:
            order = self.db_order
        d = {self.sql_map[column]: value for column, value in zip(order, row)}
        self._load_dict(d)
        if isinstance(self.categories, str):
            self.categories = loads(self.categories)

    def from_var_csv(self, row):
        from json import loads
        from csv import reader

        if self.index is not None:
            row = list(reader([row], dialect="oakvar"))[0]
            self._load_dict(dict(zip(self.csv_order[: len(row)], row)))
            self.index = int(self.index)
            if isinstance(self.categories, str):
                self.categories = loads(self.categories)
            if self.categories is None:
                self.categories = []
            if isinstance(self.hidden, str):
                self.hidden = loads(self.hidden.lower())
            if isinstance(self.filterable, str):
                self.filterable = loads(self.filterable.lower())
            if isinstance(self.hide_from_gui_filter, str):
                self.hide_from_gui_filter = loads(self.hide_from_gui_filter.lower())
            if self.link_format == "":
                self.link_format = None

    def from_json(self, sjson):
        from json import loads

        self._load_dict(loads(sjson))

    def get_json(self):
        from json import dumps

        return dumps(self.__dict__)

    def get_colinfo(self):
        return {
            "col_name": self.name,
            "col_title": self.title,
            "col_type": self.type,
            "col_cats": self.categories,
            "col_width": self.width,
            "col_desc": self.desc,
            "col_hidden": self.hidden,
            "col_ctg": self.category,
            "col_filterable": self.filterable,
            "col_hide_from_gui_filter": self.hide_from_gui_filter,
            "link_format": self.link_format,
            "col_genesummary": self.genesummary,
            "col_index": self.index,
            "table": self.table,
            "level": self.level,
        }

    def __iter__(self):  # Allows casting to dict
        for k, v in self.__dict__.items():
            yield k, v


def read_crv(fpath):
    import polars as pl

    f = open(fpath)
    c = 0
    for line in f:
        if not line.startswith("#"):
            break
        c += 1
    df = pl.read_csv(
        f,
        skip_rows=c,
        new_columns=["uid", "chrom", "pos", "pos_end", "ref_base", "alt_base"],
    )
    return df


def get_file_content_as_table(fpath: Union[Path, str], title: str, outer=None):
    from rich.table import Table

    if not outer:
        return
    table = Table(title=title, show_header=False)
    table.add_column("License")
    with open(fpath) as f:
        lines = "".join(f.readlines())
        table.add_row(lines)
    if hasattr(outer, "print"):
        outer.print(table)
    else:
        outer.write(table)


CravatReader = FileReader
CravatWriter = FileWriter



from typing import Any
from typing import Optional
from typing import List
from typing import Dict
from typing import Tuple
from re import compile
from liftover import ChainFile
from oakvar import BaseConverter

chromdict = {
    "chrx": "chrX",
    "chry": "chrY",
    "chrMT": "chrM",
    "chrMt": "chrM",
    "chr23": "chrX",
    "chr24": "chrY",
}
base_re = compile("^[ATGC]+|[-]+$")

class LinesData:
    def __init__(self, lines_data: Dict[int, List[Tuple[int, Dict[str, Any]]]]):
        self.lines_data = lines_data

def handle_genotype(variant):
    if "genotype" in variant and "." in variant["genotype"]:
        variant["genotype"] = variant["genotype"].replace(".", variant["ref_base"])

def flush_err_holder(err_holder: list, error_logger, force: bool=False):
    if len(err_holder) > 1000 or force:
        for err_line in err_holder:
            error_logger.error(err_line)
        err_holder.clear()

def _log_conversion_error(logger, error_logger, input_path: str, line_no: int, e, unique_excs: dict, err_holder: list):
    from traceback import format_exc
    from oakvar.lib.exceptions import ExpectedException
    from oakvar.lib.exceptions import NoAlternateAllele

    if isinstance(e, NoAlternateAllele):
        return
    if isinstance(e, ExpectedException):
        err_str = str(e)
    else:
        err_str = format_exc().rstrip()
    if err_str not in unique_excs:
        err_no = len(unique_excs)
        unique_excs[err_str] = err_no
        logger.error(f"Error [{err_no}]: {input_path}: {err_str}")
        err_holder.append(f"{err_no}:{line_no}\t{str(e)}")
    else:
        err_no = unique_excs[err_str]
        err_holder.append(f"{err_no}:{line_no}\t{str(e)}")
    flush_err_holder(err_holder, error_logger)

def is_chrM(wdict):
    return wdict["chrom"] == "chrM"

def perform_liftover_if_needed(variant, do_liftover: bool, do_liftover_chrM: bool, lifter, wgs_reader) -> Optional[Dict[str, Any]]:
    from copy import copy
    from oakvar.lib.util.seq import liftover_one_pos
    from oakvar.lib.util.seq import liftover

    if is_chrM(variant):
        needed = do_liftover_chrM
    else:
        needed = do_liftover
    if needed:
        prelift_wdict = copy(variant)
        crl_data = prelift_wdict
        (
            variant["chrom"],
            variant["pos"],
            variant["ref_base"],
            variant["alt_base"],
        ) = liftover(
            variant["chrom"],
            int(variant["pos"]),
            variant["ref_base"],
            variant["alt_base"],
            lifter=lifter,
            wgs_reader=wgs_reader,
        )
        converted_end = liftover_one_pos(
            variant["chrom"], variant["pos_end"], lifter=lifter
        )
        if converted_end is None:
            pos_end = ""
        else:
            pos_end = converted_end[1]
        variant["pos_end"] = pos_end
    else:
        crl_data = None
    return crl_data

def add_end_pos_if_absent(variant: dict):
    col_name = "pos_end"
    if col_name not in variant:
        ref_base = variant["ref_base"]
        ref_len = len(ref_base)
        if ref_len == 1:
            variant[col_name] = variant["pos"]
        else:
            variant[col_name] = variant["pos"] + ref_len - 1

def normalize_variant(variant):
    from oakvar.lib.util.seq import normalize_variant_left

    p, r, a = (
        int(variant["pos"]),
        variant["ref_base"],
        variant["alt_base"],
    )
    (
        new_pos,
        new_ref,
        new_alt,
    ) = normalize_variant_left("+", p, r, a)
    variant["pos"] = new_pos
    variant["ref_base"] = new_ref
    variant["alt_base"] = new_alt

def check_invalid_base(variant: dict):
    from oakvar.lib.exceptions import IgnoredVariant

    if not base_re.fullmatch(variant["ref_base"]):
        raise IgnoredVariant(f"Invalid reference base {variant['ref_base']}")
    if not base_re.fullmatch(variant["alt_base"]):
        raise IgnoredVariant(f"Invalid alternate base {variant['alt_base']}")

def handle_ref_base(variant, wgs_reader):
    from oakvar.lib.exceptions import IgnoredVariant

    if "ref_base" not in variant or variant["ref_base"] in [
        "",
        ".",
    ]:
        variant["ref_base"] = wgs_reader.get_bases(
            variant.get("chrom"), int(variant["pos"])
        ).upper()
    else:
        ref_base = variant["ref_base"]
        if ref_base == "" and variant["alt_base"] not in [
            "A",
            "T",
            "C",
            "G",
        ]:
            raise IgnoredVariant("Reference base required for non SNV")
        elif ref_base is None or ref_base == "":
            variant["ref_base"] = wgs_reader.get_bases(
                variant.get("chrom"), int(variant.get("pos"))
            )

def handle_chrom(variant):
    from oakvar.lib.exceptions import IgnoredVariant

    if not variant.get("chrom"):
        raise IgnoredVariant("No chromosome")
    if not variant.get("chrom").startswith("chr"):
        variant["chrom"] = "chr" + variant.get("chrom")
    variant["chrom"] = chromdict.get(
        variant.get("chrom"), variant.get("chrom")
    )

def is_unique_variant(variant: dict, unique_vars: dict) -> bool:
    return variant["var_no"] not in unique_vars

def handle_variant(
        variant: dict, unique_vars: dict, do_liftover: bool, do_liftover_chrM: bool, lifter, wgs_reader, line_no: int, num_valid_error_lines: Dict[str, int]
) -> Optional[Dict[str, Any]]:
    from oakvar.lib.exceptions import NoVariantError

    if variant["ref_base"] == variant["alt_base"]:
        raise NoVariantError()
    tags = variant.get("tags")
    unique = is_unique_variant(variant, unique_vars)
    if unique:
        variant["unique"] = True
        unique_vars[variant["var_no"]] = True
        handle_chrom(variant)
        handle_ref_base(variant, wgs_reader)
        check_invalid_base(variant)
        normalize_variant(variant)
        add_end_pos_if_absent(variant)
        crl_data = perform_liftover_if_needed(variant, do_liftover, do_liftover_chrM, lifter, wgs_reader)
        num_valid_error_lines["valid"] += 1
    else:
        variant["unique"] = False
        crl_data = None
    handle_genotype(variant)
    if unique:
        variant["original_line"] = line_no
        variant["tags"] = tags
    return crl_data

def handle_converted_variants(
        variants: List[Dict[str, Any]], do_liftover: bool, do_liftover_chrM: bool, lifter, wgs_reader, logger, error_logger, input_path: str, unique_excs: dict, err_holder: list, line_no: int, num_valid_error_lines: Dict[str, int]
) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
    from oakvar.lib.exceptions import IgnoredVariant

    if variants is BaseConverter.IGNORE:
        return [], []
    if not variants:
        raise IgnoredVariant("No valid alternate allele was found in any samples.")
    unique_vars = {}
    variant_l: List[Dict[str, Any]] = []
    crl_l: List[Dict[str, Any]] = []
    for variant in variants:
        try:
            crl_data = handle_variant(variant, unique_vars, do_liftover, do_liftover_chrM, lifter, wgs_reader, line_no, num_valid_error_lines)
        except Exception as e:
            _log_conversion_error(logger, error_logger, input_path, line_no, e, unique_excs, err_holder)
            continue
        variant_l.append(variant)
        if crl_data:
            crl_l.append(crl_data)
    return variant_l, crl_l

def gather_variantss(
        converter: BaseConverter, 
        lines_data: Dict[int, List[Tuple[int, Dict[str, Any]]]],
        core_num: int, 
        do_liftover: bool, 
        do_liftover_chrM: bool, 
        lifter, 
        wgs_reader, 
        logger, 
        error_logger, 
        input_path: str, 
        unique_excs: dict, 
        err_holder: list,
        num_valid_error_lines: Dict[str, int],
) -> Tuple[List[List[Dict[str, Any]]], List[List[Dict[str, Any]]]]:
    variants_l: List[List[Dict[str, Any]]] = []
    crl_l: List[List[Dict[str, Any]]] = []
    line_data = lines_data[core_num]
    for (line_no, line) in line_data:
        try:
            variants = converter.convert_line(line)
            variants_datas, crl_datas = handle_converted_variants(variants, do_liftover, do_liftover_chrM, lifter, wgs_reader, logger, error_logger, input_path, unique_excs, err_holder, line_no, num_valid_error_lines)
            if not variants_datas:
                continue
            variants_l.append(variants_datas)
            crl_l.append(crl_datas)
        except KeyboardInterrupt:
            raise
        except Exception as e:
            _log_conversion_error(logger, error_logger, input_path, line_no, e, unique_excs, err_holder)
            num_valid_error_lines["error"] += 1
    return variants_l, crl_l

def gather_variantss_wrapper(args):
    return gather_variantss(*args)

class MasterConverter(object):
    def __init__(
        self,
        inputs: List[str] = [],
        input_format: Optional[str] = None,
        name: Optional[str] = None,
        output_dir: Optional[str] = None,
        genome: Optional[str] = None,
        serveradmindb=None,
        conf: Dict = {},
        module_options: Dict = {},
        input_encoding=None,
        ignore_sample: bool=False,
        mp: int=1,
        outer=None,
    ):
        from re import compile
        from oakvar import get_wgs_reader
        from oakvar.lib.exceptions import ExpectedException

        self.logger = None
        self.crv_writer = None
        self.crs_writer = None
        self.crm_writer = None
        self.crl_writer = None
        self.converters = {}
        self.available_input_formats = []
        self.input_paths = None
        self.input_dir = None
        self.input_path_dict = {}
        self.input_path_dict2 = {}
        self.input_file_handles: Dict[str, str] = {}
        self.output_base_fname: Optional[str] = None
        self.error_logger = None
        self.unique_excs: Dict[str, int] = {}
        self.err_holder = []
        self.wpath = None
        self.crm_path = None
        self.crs_path = None
        self.crl_path = None
        self.do_liftover = None
        self.do_liftover_chrM = None
        self.lifter = None
        self.module_options = None
        self.given_input_assembly: Optional[str] = None
        self.converter_by_input_path: Dict[str, Optional[BaseConverter]] = {}
        self.extra_output_columns = []
        self.file_num_valid_variants = 0
        self.file_error_lines = 0
        self.total_num_valid_variants = 0
        self.total_error_lines = 0
        self.fileno = 0
        self.ignore_sample = ignore_sample
        self.total_num_converted_variants = 0
        self.input_formats: List[str] = []
        self.genome_assemblies: List[str] = []
        self.base_re = compile("^[ATGC]+|[-]+$")
        self.last_var: str = ""
        self.unique_vars: dict = {}
        self.chromdict = {
            "chrx": "chrX",
            "chry": "chrY",
            "chrMT": "chrM",
            "chrMt": "chrM",
            "chr23": "chrX",
            "chr24": "chrY",
        }
        if not inputs:
            raise ExpectedException("Input files are not given.")
        self.inputs = inputs
        self.format = input_format
        self.name = name
        self.output_dir = output_dir
        self.genome = genome
        self.parse_inputs()
        self.parse_output_dir()
        self.given_input_assembly = genome
        self.conf: Dict = {}
        self.module_options = module_options
        if conf:
            self.conf.update(conf)
        self.serveradmindb = serveradmindb
        self.input_encoding = input_encoding
        self.outer = outer
        self.setup_logger()
        self.wgs_reader = get_wgs_reader(assembly="hg38")
        self.time_error_written: float = 0
        self.mp = mp

    def get_genome_assembly(self, converter) -> str:
        from oakvar.lib.system.consts import default_assembly_key
        from oakvar.lib.exceptions import NoGenomeException
        from oakvar.lib.system import get_user_conf

        if self.given_input_assembly:
            return self.given_input_assembly
        input_assembly = getattr(converter, "input_assembly", None)
        if input_assembly:
            return input_assembly
        user_conf = get_user_conf() or {}
        genome_assembly = user_conf.get(default_assembly_key, None)
        if genome_assembly:
            return genome_assembly
        raise NoGenomeException()

    def set_do_liftover(self, genome_assembly, converter, input_path):
        self.do_liftover = genome_assembly != "hg38"
        if hasattr(converter, "get_do_liftover_chrM"):
            self.do_liftover_chrM = converter.get_do_liftover_chrM(
                genome_assembly, input_path, self.do_liftover
            )
        else:
            self.do_liftover_chrM = self.do_liftover
        if self.logger:
            self.logger.info(f"liftover needed: {self.do_liftover}")
            self.logger.info(f"liftover for chrM needed: {self.do_liftover_chrM}")

    def setup_lifter(self, genome_assembly) -> Optional[ChainFile]:
        from oakvar.lib.util.seq import get_lifter

        self.lifter = get_lifter(source_assembly=genome_assembly)

    def parse_inputs(self):
        from pathlib import Path

        self.input_paths = []
        self.input_paths = [
            str(Path(x).absolute()) for x in self.inputs if x != "-"
        ]
        self.input_dir = str(Path(self.input_paths[0]).parent)
        for i in range(len(self.input_paths)):
            self.input_path_dict[i] = self.input_paths[i]
            self.input_path_dict2[self.input_paths[i]] = i

    def parse_output_dir(self):
        from pathlib import Path
        from os import makedirs

        if not self.output_dir:
            self.output_dir = self.input_dir
        if not self.output_dir:
            raise
        if not (Path(self.output_dir).exists()):
            makedirs(self.output_dir)
        self.output_base_fname: Optional[str] = self.name
        if not self.output_base_fname:
            if not self.input_paths:
                raise
            self.output_base_fname = Path(self.input_paths[0]).name

    def get_file_object_for_input_path(self, input_path: str):
        from pathlib import Path
        from oakvar.lib.util.util import detect_encoding

        suffix = Path(input_path).suffix
        # TODO: Remove the hardcoding.
        if suffix in [".parquet"]:
            encoding = ""
        else:
            if self.input_encoding:
                encoding = self.input_encoding
            else:
                if self.logger:
                    self.logger.info(f"detecting encoding of {input_path}")
                encoding = detect_encoding(input_path)
        if self.logger:
            self.logger.info(f"encoding: {input_path} {encoding}")
        return encoding

    def collect_input_file_handles(self):
        if not self.input_paths:
            raise
        for input_path in self.input_paths:
            encoding = self.get_file_object_for_input_path(input_path)
            self.input_file_handles[input_path] = encoding

    def setup(self):
        self.collect_converters()
        self.collect_input_file_handles()
        self.check_input_format()
        self.collect_converter_by_input()
        self.collect_extra_output_columns()
        self.open_output_files()

    def setup_logger(self):
        from logging import getLogger

        self.logger = getLogger("oakvar.converter")
        self.error_logger = getLogger("err.converter")

    def collect_converters(self):
        from oakvar.lib.module.local import get_local_module_infos_of_type
        from oakvar.lib.util.util import load_class

        for module_name, module_info in get_local_module_infos_of_type(
            "converter"
        ).items():
            try:
                cls = load_class(module_info.script_path)
                if cls is None:
                    continue
                if self.module_options is not None:
                    module_conf = self.module_options.get(module_name)
                else:
                    module_conf = None
                converter = cls(ignore_sample=self.ignore_sample, module_conf=module_conf)
            except Exception:
                if self.logger:
                    self.logger.error(f"Skipping {module_name} as it could not be loaded.")
                continue
            # TODO: backward compatibility
            converter.output_dir = None
            converter.run_name = None
            converter.module_name = module_name
            converter.name = module_info.name
            converter.version = module_info.version
            converter.conf = module_info.conf
            converter.script_path = module_info.script_path
            format_name = module_info.conf.get("format_name")
            # end of backward compatibility
            if format_name:
                converter.format_name = format_name
            elif not hasattr(converter, "format_name"):
                converter.format_name = module_name.split("-")[0]
            if not converter.format_name:
                if self.logger:
                    self.logger.info(
                        f"Skipping {module_info.name} as it does not "
                        + "have format_name defined."
                    )
                continue
            converter.module_name = module_info.name
            if converter.format_name not in self.converters:
                self.converters[converter.format_name] = converter
            else:
                if self.outer:
                    self.outer.write(
                        f"{module_info.name} is skipped because "
                        + f"{converter.format_name} is already handled by "
                        + f"{self.converters[converter.format_name].name}."
                    )
                continue
        self.available_input_formats = list(self.converters.keys())

    def collect_converter_by_input(self):
        if not self.input_paths:
            return
        for input_path in self.input_paths:
            converter = self.get_converter_for_input_file(input_path)
            self.converter_by_input_path[input_path] = converter

    def collect_extra_output_columns(self):
        for converter in self.converter_by_input_path.values():
            if not converter:
                continue
            if converter.conf:
                extra_output_columns = converter.conf.get("extra_output_columns")
                if not extra_output_columns:
                    continue
                for col in extra_output_columns:
                    self.extra_output_columns.append(col)

    def check_input_format(self):
        from oakvar.lib.exceptions import InvalidInputFormat

        if self.format and self.format not in self.available_input_formats:
            raise InvalidInputFormat(self.format)

    def get_converter_for_input_file(self, input_path) -> Optional[BaseConverter]:
        converter: Optional[BaseConverter] = None
        if self.format:
            if self.format in self.converters:
                converter = self.converters[self.format]
        else:
            for check_converter in self.converters.values():
                try:
                    check_success = check_converter.check_format(input_path)
                except Exception:
                    import traceback

                    if self.error_logger:
                        self.error_logger.error(traceback.format_exc())
                    check_success = False
                if check_success:
                    converter = check_converter
                    break
        if converter:
            if self.logger:
                self.logger.info(f"Using {converter.module_name} for {input_path}")
            return converter
        return None

    def set_converter_properties(self, converter):
        from oakvar.lib.exceptions import SetupError
        from oakvar.lib.module.local import get_module_code_version

        if self.conf is None:
            raise SetupError()
        converter.output_dir = self.output_dir
        converter.run_name = self.output_base_fname
        module_name = converter.module_name
        converter.version = get_module_code_version(converter.module_name)
        if module_name in self.conf:
            if hasattr(converter, "conf") is False:
                converter.conf = {}
            converter.conf.update(self.conf[module_name])

    def setup_crv_writer(self):
        from pathlib import Path
        from oakvar.lib.util.util import get_crv_def
        from oakvar.lib.util.inout import FileWriter
        from oakvar.lib.consts import crv_idx
        from oakvar.lib.consts import STANDARD_INPUT_FILE_SUFFIX

        if not self.output_dir or not self.output_base_fname:
            raise
        crv_def = get_crv_def()
        self.wpath = Path(self.output_dir) / (
            self.output_base_fname + STANDARD_INPUT_FILE_SUFFIX
        )
        self.crv_writer = FileWriter(self.wpath)
        self.crv_writer.add_columns(crv_def)
        self.crv_writer.write_definition()
        for index_columns in crv_idx:
            self.crv_writer.add_index(index_columns)

    def setup_crs_writer(self):
        from pathlib import Path
        from oakvar.lib.util.util import get_crs_def
        from oakvar.lib.util.inout import FileWriter
        from oakvar.lib.consts import crs_idx
        from oakvar.lib.consts import SAMPLE_FILE_SUFFIX

        if not self.output_dir or not self.output_base_fname:
            raise
        crs_def = get_crs_def()
        self.crs_path = Path(self.output_dir) / (
            self.output_base_fname + SAMPLE_FILE_SUFFIX
        )
        self.crs_writer = FileWriter(self.crs_path)
        self.crs_writer.add_columns(crs_def)
        self.crs_writer.add_columns(self.extra_output_columns)
        self.crs_writer.write_definition()
        for index_columns in crs_idx:
            self.crs_writer.add_index(index_columns)

    def setup_crm_writer(self):
        from pathlib import Path
        from oakvar.lib.util.util import get_crm_def
        from oakvar.lib.util.inout import FileWriter
        from oakvar.lib.consts import crm_idx
        from oakvar.lib.consts import MAPPING_FILE_SUFFIX

        if not self.output_dir or not self.output_base_fname:
            raise
        crm_def = get_crm_def()
        self.crm_path = Path(self.output_dir) / (
            self.output_base_fname + MAPPING_FILE_SUFFIX
        )
        self.crm_writer = FileWriter(self.crm_path)
        self.crm_writer.add_columns(crm_def)
        self.crm_writer.write_definition()
        for index_columns in crm_idx:
            self.crm_writer.add_index(index_columns)
        self.crm_writer.write_input_paths(self.input_path_dict)

    def setup_crl_writer(self):
        from pathlib import Path
        from oakvar.lib.util.util import get_crl_def
        from oakvar.lib.util.inout import FileWriter
        from oakvar.lib.consts import VARIANT_LEVEL_OUTPUT_SUFFIX

        if not self.output_dir or not self.output_base_fname:
            raise
        crl_def = get_crl_def()
        self.crl_path = (
            Path(self.output_dir)
            / f"{self.output_base_fname}.original_input{VARIANT_LEVEL_OUTPUT_SUFFIX}"
        )
        self.crl_writer = FileWriter(self.crl_path)
        self.crl_writer.add_columns(crl_def)
        self.crl_writer.write_definition()
        self.crl_writer.write_names("original_input", "Original Input", "")

    def open_output_files(self):
        from oakvar.lib.exceptions import SetupError

        if not self.output_base_fname or not self.output_dir:
            raise SetupError()
        self.setup_crv_writer()
        self.setup_crs_writer()
        self.setup_crm_writer()
        self.setup_crl_writer()

    def log_input_and_genome_assembly(self, input_path, genome_assembly, converter):
        if not self.logger:
            return
        self.logger.info(f"input file: {input_path}")
        self.logger.info("input format: %s" % converter.format_name)
        self.logger.info(f"genome_assembly: {genome_assembly}")

    def setup_file(self, input_path: str) -> BaseConverter:
        from logging import getLogger
        from oakvar.lib.util.util import log_module
        from oakvar.lib.exceptions import NoConverterFound

        encoding = self.input_file_handles[input_path]
        converter = self.converter_by_input_path[input_path]
        if not converter:
            raise NoConverterFound(input_path)
        if not converter.module_name or not converter.format_name:
            raise NoConverterFound(input_path)
        self.input_formats.append(converter.format_name)
        self.fileno += 1
        self.set_converter_properties(converter)
        log_module(converter, self.logger)
        self.error_logger = getLogger("err." + converter.module_name)
        converter.input_path = input_path
        converter.input_paths = self.input_paths
        converter.setup(input_path, encoding=encoding)
        genome_assembly = self.get_genome_assembly(converter)
        self.genome_assemblies.append(genome_assembly)
        self.log_input_and_genome_assembly(input_path, genome_assembly, converter)
        self.set_do_liftover(genome_assembly, converter, input_path)
        if self.do_liftover or self.do_liftover_chrM:
            self.setup_lifter(genome_assembly)
        return converter

    def handle_chrom(self, variant):
        from oakvar.lib.exceptions import IgnoredVariant

        if not variant.get("chrom"):
            raise IgnoredVariant("No chromosome")
        if not variant.get("chrom").startswith("chr"):
            variant["chrom"] = "chr" + variant.get("chrom")
        variant["chrom"] = self.chromdict.get(
            variant.get("chrom"), variant.get("chrom")
        )

    def handle_ref_base(self, variant):
        from oakvar.lib.exceptions import IgnoredVariant

        if "ref_base" not in variant or variant["ref_base"] in [
            "",
            ".",
        ]:
            if not self.wgs_reader:
                raise
            variant["ref_base"] = self.wgs_reader.get_bases(
                variant.get("chrom"), int(variant["pos"])
            ).upper()
        else:
            ref_base = variant["ref_base"]
            if ref_base == "" and variant["alt_base"] not in [
                "A",
                "T",
                "C",
                "G",
            ]:
                raise IgnoredVariant("Reference base required for non SNV")
            elif ref_base is None or ref_base == "":
                if not self.wgs_reader:
                    raise
                variant["ref_base"] = self.wgs_reader.get_bases(
                    variant.get("chrom"), int(variant.get("pos"))
                )

    def handle_genotype(self, variant):
        if "genotype" in variant and "." in variant["genotype"]:
            variant["genotype"] = variant["genotype"].replace(".", variant["ref_base"])

    def check_invalid_base(self, variant: dict):
        from oakvar.lib.exceptions import IgnoredVariant

        if not self.base_re.fullmatch(variant["ref_base"]):
            raise IgnoredVariant(f"Invalid reference base {variant['ref_base']}")
        if not self.base_re.fullmatch(variant["alt_base"]):
            raise IgnoredVariant(f"Invalid alternate base {variant['alt_base']}")

    def normalize_variant(self, variant):
        from oakvar.lib.util.seq import normalize_variant_left

        p, r, a = (
            int(variant["pos"]),
            variant["ref_base"],
            variant["alt_base"],
        )
        (
            new_pos,
            new_ref,
            new_alt,
        ) = normalize_variant_left("+", p, r, a)
        variant["pos"] = new_pos
        variant["ref_base"] = new_ref
        variant["alt_base"] = new_alt

    def add_end_pos_if_absent(self, variant: dict):
        col_name = "pos_end"
        if col_name not in variant:
            ref_base = variant["ref_base"]
            ref_len = len(ref_base)
            if ref_len == 1:
                variant[col_name] = variant["pos"]
            else:
                variant[col_name] = variant["pos"] + ref_len - 1

    def run(self):
        from pathlib import Path
        from multiprocessing.pool import ThreadPool
        from sys import platform as sysplatform
        #import time
        from oakvar.lib.util.run import update_status

        if not self.input_paths or not self.logger:
            raise
        update_status(
            "started converter", logger=self.logger, serveradmindb=self.serveradmindb
        )
        self.setup()
        self.set_variables_pre_run()
        if not self.crv_writer:
            raise ValueError("No crv_writer")
        if not self.crs_writer:
            raise ValueError("No crs_writer")
        if not self.crm_writer:
            raise ValueError("No crm_writer")
        if not self.crl_writer:
            raise ValueError("No crl_writer")
        if sysplatform == "win32": # TODO: Remove after releasing Rust-based vcf-converter.
            batch_size: int = 10000
            num_pool = 1
        else:
            batch_size: int = 2500
            num_pool = 4
        uid = 1
        pool = ThreadPool(num_pool)
        for input_path in self.input_paths:
            self.input_fname = Path(input_path).name
            fileno = self.input_path_dict2[input_path]
            converter = self.setup_file(input_path)
            self.file_num_valid_variants = 0
            self.file_error_lines = 0
            self.num_valid_error_lines = {"valid": 0, "error": 0}
            start_line_pos: int = 1
            start_line_no: int = start_line_pos
            round_no: int = 0
            #stime = time.time()
            while True:
                #ctime = time.time()
                lines_data, immature_exit = converter.get_variant_lines(input_path, num_pool, start_line_no, batch_size)
                args = [
                    (
                        converter, 
                        lines_data,
                        core_num, 
                        self.do_liftover, 
                        self.do_liftover_chrM, 
                        self.lifter, 
                        self.wgs_reader, 
                        self.logger, 
                        self.error_logger, 
                        input_path, 
                        self.unique_excs, 
                        self.err_holder,
                        self.num_valid_error_lines,
                    ) for core_num in range(num_pool)
                ]
                results = pool.map(gather_variantss_wrapper, args)
                lines_data = None
                for result in results:
                    variants_l, crl_l = result
                    for i in range(len(variants_l)):
                        variants = variants_l[i]
                        crl_data = crl_l[i]
                        if len(variants) == 0:
                            continue
                        if crl_data:
                            for variant, crl in zip(variants, crl_data):
                                uid_var = uid + variant["var_no"]
                                variant["uid"] = uid_var
                                crl["uid"] = uid_var
                                if variant["unique"]:
                                    self.crv_writer.write_data(variant)
                                    variant["fileno"] = fileno
                                    self.crm_writer.write_data(variant)
                                    converter.write_extra_info(variant)
                                self.crs_writer.write_data(variant)
                                self.crl_writer.write_data(crl)
                        else:
                            for variant in variants:
                                uid_var = uid + variant["var_no"]
                                variant["uid"] = uid_var
                                if variant["unique"]:
                                    self.crv_writer.write_data(variant)
                                    variant["fileno"] = fileno
                                    self.crm_writer.write_data(variant)
                                    converter.write_extra_info(variant)
                                self.crs_writer.write_data(variant)
                        uid += max([v["var_no"] for v in variants]) + 1
                    variants_l = None
                    crl_l = None
                if not immature_exit:
                    break
                start_line_no += batch_size * num_pool
                round_no += 1
                status = (
                    f"Running Converter ({self.input_fname}): line {start_line_no - 1}"
                )
                update_status(
                    status, logger=self.logger, serveradmindb=self.serveradmindb
                )
            self.logger.info(
                f"{input_path}: number of valid variants: {self.num_valid_error_lines['valid']}"
            )
            self.logger.info(f"{input_path}: number of lines skipped due to errors: {self.num_valid_error_lines['error']}")
            self.total_num_converted_variants += self.num_valid_error_lines["valid"]
            self.total_num_valid_variants += self.num_valid_error_lines["valid"]
            self.total_error_lines += self.num_valid_error_lines["error"]
        flush_err_holder(self.err_holder, self.error_logger, force=True)
        self.close_output_files()
        self.end()
        self.log_ending()
        ret = {
            "total_lnum": self.total_num_converted_variants,
            "write_lnum": self.total_num_valid_variants,
            "error_lnum": self.total_error_lines,
            "input_formats": self.input_formats,
            "assemblies": self.genome_assemblies,
        }
        return ret

    def set_variables_pre_run(self):
        from time import time

        self.start_time = time()
        self.total_num_converted_variants = 0
        self.uid = 0

    def log_ending(self):
        from time import time, asctime, localtime
        from oakvar.lib.util.run import update_status

        if not self.logger:
            raise
        self.logger.info(
            "total number of valid variants: {}".format(
                self.total_num_converted_variants
            )
        )
        self.logger.info("total number of lines skipped due to errors: %d" % self.total_error_lines)
        end_time = time()
        self.logger.info("finished: %s" % asctime(localtime(end_time)))
        runtime = round(end_time - self.start_time, 3)
        self.logger.info("runtime: %s" % runtime)
        status = "finished Converter"
        update_status(status, logger=self.logger, serveradmindb=self.serveradmindb)

    def perform_liftover_if_needed(self, variant) -> Optional[Dict[str, Any]]:
        from copy import copy
        from oakvar.lib.util.seq import liftover_one_pos
        from oakvar.lib.util.seq import liftover

        assert self.crl_writer is not None
        if self.is_chrM(variant):
            needed = self.do_liftover_chrM
        else:
            needed = self.do_liftover
        if needed:
            prelift_wdict = copy(variant)
            #prelift_wdict["uid"] = self.uid
            #self.crl_writer.write_data(prelift_wdict)
            crl_data = prelift_wdict
            (
                variant["chrom"],
                variant["pos"],
                variant["ref_base"],
                variant["alt_base"],
            ) = liftover(
                variant["chrom"],
                int(variant["pos"]),
                variant["ref_base"],
                variant["alt_base"],
                lifter=self.lifter,
                wgs_reader=self.wgs_reader,
            )
            converted_end = liftover_one_pos(
                variant["chrom"], variant["pos_end"], lifter=self.lifter
            )
            if converted_end is None:
                pos_end = ""
            else:
                pos_end = converted_end[1]
            variant["pos_end"] = pos_end
        else:
            crl_data = None
        return crl_data

    def is_chrM(self, wdict):
        return wdict["chrom"] == "chrM"

    def close_output_files(self):
        if self.crv_writer is not None:
            self.crv_writer.close()
        if self.crm_writer is not None:
            self.crm_writer.close()
        if self.crs_writer is not None:
            self.crs_writer.close()

    def end(self):
        pass



from . import cli_entry
from . import cli_func


@cli_entry
def cli_license(args):
    return license(args)


@cli_func
def license(args, __name__="license"):
    from ..api import license

    return license(**args)


def get_parser_ov_license():
    from argparse import ArgumentParser

    parser_ov_license = ArgumentParser(
        description="Shows license information.",
    )
    parser_ov_license.set_defaults(func=cli_license)
    return parser_ov_license
